"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __values = (this && this.__values) || function(o) {
    var s = typeof Symbol === "function" && Symbol.iterator, m = s && o[s], i = 0;
    if (m) return m.call(o);
    if (o && typeof o.length === "number") return {
        next: function () {
            if (o && i >= o.length) o = void 0;
            return { value: o && o[i++], done: !o };
        }
    };
    throw new TypeError(s ? "Object is not iterable." : "Symbol.iterator is not defined.");
};
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ArgsParser = exports.ParsedArgumentsCollection = void 0;
var exceptions_1 = require("./exceptions");
var utils_1 = require("./utils/utils");
var help_1 = require("./utils/help");
var validation_1 = require("./utils/validation");
var cast_1 = require("./utils/cast");
/**
 * A collection of parsed arguments.
 *
 * @category Classes
 */
var ParsedArgumentsCollection = /** @class */ (function () {
    function ParsedArgumentsCollection() {
        /**
         * A collection of parsed positional arguments.
         */
        this.positionalArgs = {};
        /**
         * A collection of parsed optional arguments.
         */
        this.optionArgs = {};
    }
    Object.defineProperty(ParsedArgumentsCollection.prototype, "positional", {
        /**
         * Returns all arguments as a record.
         */
        get: function () {
            return __assign({}, this.positionalArgs);
        },
        enumerable: false,
        configurable: true
    });
    Object.defineProperty(ParsedArgumentsCollection.prototype, "options", {
        /**
         * Returns all arguments as a record.
         */
        get: function () {
            return __assign({}, this.optionArgs);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Adds a new argument to the collection.
     * @param name - The name of the argument.
     * @param value - The value of the argument.
     * @returns The instance of ParsedArgumentsCollection for chaining.
     */
    ParsedArgumentsCollection.prototype.add = function (name, value) {
        var formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            this.positionalArgs[formattedName] = value;
        }
        else {
            this.optionArgs[formattedName] = value;
        }
        return this;
    };
    /**
     * Retrieves an argument value.
     * @param name - The name of the argument.
     * @returns The value of the argument, or undefined if not found.
     */
    ParsedArgumentsCollection.prototype.get = function (name) {
        var _a, _b;
        var formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            return ((_a = this.positional[formattedName]) !== null && _a !== void 0 ? _a : undefined);
        }
        return ((_b = this.optionArgs[formattedName]) !== null && _b !== void 0 ? _b : undefined);
    };
    /**
     * Checks if an argument exists in the collection.
     * @param name - The name of the argument.
     * @returns True if the argument exists, otherwise false.
     */
    ParsedArgumentsCollection.prototype.has = function (name) {
        var _a, _b;
        var formattedName = this.formatName(name);
        if (this.isPositional(name)) {
            return ((_a = this.positional[formattedName]) !== null && _a !== void 0 ? _a : undefined) !== undefined;
        }
        return ((_b = this.optionArgs[formattedName]) !== null && _b !== void 0 ? _b : undefined) !== undefined;
    };
    /**
     * Formats the argument name by removing leading dashes.
     * @param name - The argument name.
     * @returns The formatted name.
     */
    ParsedArgumentsCollection.prototype.formatName = function (name) {
        return name.replace(/^--/, '');
    };
    /**
     * Checks if the argument is positional.
     * @param name - The argument name.
     * @returns True if the argument is positional, otherwise false.
     */
    ParsedArgumentsCollection.prototype.isPositional = function (name) {
        return !name.startsWith('--');
    };
    return ParsedArgumentsCollection;
}());
exports.ParsedArgumentsCollection = ParsedArgumentsCollection;
/**
 * Parser for command-line arguments.
 *
 * @category Classes
 *
 * @example
 * ```
 * const parser = new ArgsParser([
 *    {
 *        name: 'action',
 *        description: 'Action to perform',
 *        type: 'string',
 *    },
 *    {
 *        name: '--flag',
 *        description: 'Flag',
 *        alias: '-f',
 *        type: 'boolean',
 *        const: true,
 *        default: false,
 *    },
 *    {
 *        name: '--values',
 *        description: 'Values',
 *        alias: '-v',
 *        type: 'string',
 *        nargs: '*',
 *        default: [],
 *    }
 * ]);
 *
 * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
 * const parsedArgs = parser.parse(argv);
 *
 * console.log(parsedArgs.positional); // { action: 'make' }
 * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
 * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
 * ```
 */
var ArgsParser = /** @class */ (function () {
    /**
     * Constructs an instance of ArgsParser with the provided argument configurations.
     * @param args - An array of argument configurations.
     */
    function ArgsParser(args) {
        var e_1, _a;
        if (args === void 0) { args = []; }
        /**
         * Maps argument names to their configuration.
         */
        this.argsMap = new Map();
        /**
         * Maps argument aliases to their corresponding names.
         */
        this.aliasMap = new Map();
        /**
         * Tracks used argument names and aliases to prevent duplicates.
         */
        this.usedArgs = new Set();
        try {
            for (var args_1 = __values(args), args_1_1 = args_1.next(); !args_1_1.done; args_1_1 = args_1.next()) {
                var arg = args_1_1.value;
                this.addArgument(arg);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (args_1_1 && !args_1_1.done && (_a = args_1.return)) _a.call(args_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    }
    Object.defineProperty(ArgsParser.prototype, "help", {
        /**
         * Retrieves the help message.
         */
        get: function () {
            var e_2, _a, e_3, _b;
            var positionalArgs = this.getPositionalArguments();
            var optionArgs = this.getOptionArguments();
            var positionalArgsRows = [];
            try {
                for (var positionalArgs_1 = __values(positionalArgs), positionalArgs_1_1 = positionalArgs_1.next(); !positionalArgs_1_1.done; positionalArgs_1_1 = positionalArgs_1.next()) {
                    var arg = positionalArgs_1_1.value;
                    positionalArgsRows.push.apply(positionalArgsRows, __spreadArray([], __read((0, help_1.formatArgHelp)(arg)), false));
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (positionalArgs_1_1 && !positionalArgs_1_1.done && (_a = positionalArgs_1.return)) _a.call(positionalArgs_1);
                }
                finally { if (e_2) throw e_2.error; }
            }
            var optionalArgsRows = [];
            try {
                for (var optionArgs_1 = __values(optionArgs), optionArgs_1_1 = optionArgs_1.next(); !optionArgs_1_1.done; optionArgs_1_1 = optionArgs_1.next()) {
                    var arg = optionArgs_1_1.value;
                    optionalArgsRows.push.apply(optionalArgsRows, __spreadArray([], __read((0, help_1.formatArgHelp)(arg)), false));
                }
            }
            catch (e_3_1) { e_3 = { error: e_3_1 }; }
            finally {
                try {
                    if (optionArgs_1_1 && !optionArgs_1_1.done && (_b = optionArgs_1.return)) _b.call(optionArgs_1);
                }
                finally { if (e_3) throw e_3.error; }
            }
            var result = [];
            if (positionalArgsRows.length > 0) {
                result.push.apply(result, __spreadArray([['Positional arguments:'],
                    ['']], __read((0, help_1.tabTableRows)(positionalArgsRows)), false));
            }
            if (optionalArgsRows.length > 0) {
                result.push.apply(result, __spreadArray([['Options:'],
                    ['']], __read((0, help_1.tabTableRows)(optionalArgsRows)), false));
            }
            return (0, help_1.convertToTable)(result, 1);
        },
        enumerable: false,
        configurable: true
    });
    /**
     * Adds the help message for the arguments to the parser.
     * @returns The updated parser.
     */
    ArgsParser.prototype.addHelp = function () {
        return this.addArgument({
            name: '--help',
            alias: '-h',
            description: 'Show help',
            type: 'boolean',
            const: true,
            default: false,
        });
    };
    /**
     * Adds a new argument configuration to the parser.
     * @param config - The argument configuration.
     * @returns The instance of ArgsParser for chaining.
     *
     * @example
     * ```
     * const parser = new ArgsParser();
     * parser.addArgument({
     *   name: '--flag',
     *   alias: '-f',
     *   type: 'boolean',
     *   const: true,
     *   default: false,
     * });
     * ```
     */
    ArgsParser.prototype.addArgument = function (config) {
        (0, validation_1.validateArgConfig)(config, this.usedArgs);
        this.argsMap.set(config.name, this.extendArgConfig(config));
        this.usedArgs.add(config.name);
        if (config.alias !== undefined) {
            this.aliasMap.set(config.alias, config.name);
            this.usedArgs.add(config.alias);
        }
        return this;
    };
    /**
     * Parses the given argument string and returns a collection of parsed arguments.
     * @param argv - The argument string.
     * @returns A ParsedArgumentsCollection containing the parsed arguments.
     *
     * @example
     * ```
     * const parser = new ArgsParser([
     *    {
     *        name: 'action',
     *        description: 'Action to perform',
     *        type: 'string',
     *    },
     *    {
     *        name: '--flag',
     *        description: 'Flag',
     *        alias: '-f',
     *        type: 'boolean',
     *        const: true,
     *        default: false,
     *    },
     *    {
     *        name: '--values',
     *        description: 'Values',
     *        alias: '-v',
     *        type: 'string',
     *        nargs: '*',
     *        default: [],
     *    }
     * ]);
     *
     * const argv = ['make', '--flag', '-v', 'a', 'b', 'c'];
     * const parsedArgs = parser.parse(argv);
     *
     * console.log(parsedArgs.positional); // { action: 'make' }
     * console.log(parsedArgs.optional); // { flag: true, values: ['a', 'b', 'c'] }
     * console.log(parsedArgs.get('--values')); // ['a', 'b', 'c']
     * ```
     */
    ArgsParser.prototype.parse = function (argv) {
        var e_4, _a, e_5, _b, e_6, _c;
        var positionalArgs = this.getPositionalArguments();
        var optionalArgs = new Set(this.getOptionArguments());
        var _d = __read((0, utils_1.parseArgsArray)(argv), 2), parsedPositional = _d[0], parsedOptions = _d[1];
        parsedPositional.reverse();
        var result = new ParsedArgumentsCollection();
        try {
            for (var positionalArgs_2 = __values(positionalArgs), positionalArgs_2_1 = positionalArgs_2.next(); !positionalArgs_2_1.done; positionalArgs_2_1 = positionalArgs_2.next()) {
                var argConfig = positionalArgs_2_1.value;
                (0, validation_1.checkEnoughPositionalValues)(parsedPositional, argConfig);
                var toReadCount = this.getArgsCountToRead(argConfig, parsedPositional.length);
                var value = [];
                for (var i = 0; i < toReadCount; ++i) {
                    value.push(parsedPositional.pop());
                }
                result.add(argConfig.name, this.processArgValue(value, argConfig, value.length > 0));
            }
        }
        catch (e_4_1) { e_4 = { error: e_4_1 }; }
        finally {
            try {
                if (positionalArgs_2_1 && !positionalArgs_2_1.done && (_a = positionalArgs_2.return)) _a.call(positionalArgs_2);
            }
            finally { if (e_4) throw e_4.error; }
        }
        (0, validation_1.checkAllPositionalValuesUsed)(parsedPositional);
        try {
            for (var _e = __values(Object.entries(parsedOptions)), _f = _e.next(); !_f.done; _f = _e.next()) {
                var _g = __read(_f.value, 2), key = _g[0], value = _g[1];
                var argConfig = this.getArgConfig(key);
                result.add(argConfig.name, this.processArgValue(value, argConfig, true));
                optionalArgs.delete(argConfig);
            }
        }
        catch (e_5_1) { e_5 = { error: e_5_1 }; }
        finally {
            try {
                if (_f && !_f.done && (_b = _e.return)) _b.call(_e);
            }
            finally { if (e_5) throw e_5.error; }
        }
        try {
            for (var optionalArgs_1 = __values(optionalArgs), optionalArgs_1_1 = optionalArgs_1.next(); !optionalArgs_1_1.done; optionalArgs_1_1 = optionalArgs_1.next()) {
                var argConfig = optionalArgs_1_1.value;
                result.add(argConfig.name, this.processArgValue([], argConfig, false));
            }
        }
        catch (e_6_1) { e_6 = { error: e_6_1 }; }
        finally {
            try {
                if (optionalArgs_1_1 && !optionalArgs_1_1.done && (_c = optionalArgs_1.return)) _c.call(optionalArgs_1);
            }
            finally { if (e_6) throw e_6.error; }
        }
        return result;
    };
    ArgsParser.prototype.processArgValue = function (value, argConfig, isset) {
        var validator = (0, validation_1.createValueValidator)(argConfig);
        var caster = (0, cast_1.createValueCaster)(argConfig);
        validator.validateBeforeCast(value, isset);
        var result = caster.cast(value, isset);
        validator.validateAfterCast(result);
        return result;
    };
    ArgsParser.prototype.getArgsCountToRead = function (nargsConfig, totalCount) {
        if (!nargsConfig.multiple) {
            return Math.min(1, totalCount);
        }
        if (nargsConfig.valuesCount !== undefined) {
            return Math.min(nargsConfig.valuesCount, totalCount);
        }
        return totalCount;
    };
    /**
     * Retrieves the argument configuration for a given key.
     * @param key - The argument key.
     * @returns The corresponding ArgConfigExtended.
     * @throws ArgumentNameError if the argument key is unknown.
     */
    ArgsParser.prototype.getArgConfig = function (key) {
        var _a;
        var config = this.argsMap.get((_a = this.aliasMap.get(key)) !== null && _a !== void 0 ? _a : key);
        if (config === undefined) {
            throw new exceptions_1.ArgumentNameError("Unknown argument ".concat(key, "."));
        }
        return config;
    };
    /**
     * Retrieves the positional arguments.
     * @returns An array of positional argument configurations.
     */
    ArgsParser.prototype.getPositionalArguments = function () {
        return __spreadArray([], __read(this.argsMap.values()), false).filter(function (x) { return x.positional; });
    };
    /**
     * Retrieves the optional arguments.
     * @returns An array of optional argument configurations.
     */
    ArgsParser.prototype.getOptionArguments = function () {
        return __spreadArray([], __read(this.argsMap.values()), false).filter(function (x) { return !x.positional; });
    };
    ArgsParser.prototype.extendArgConfig = function (config) {
        return __assign(__assign({}, config), this.buildNArgsConfig(config));
    };
    ArgsParser.prototype.buildNArgsConfig = function (config) {
        var _a;
        var positional = !config.name.startsWith('--');
        var multiple = config.nargs === '*' || config.nargs === '+' || typeof config.nargs === 'number';
        var required = (_a = config.required) !== null && _a !== void 0 ? _a : (config.nargs !== '*' && config.nargs !== '?' && config.default === undefined);
        var allowEmpty = config.nargs === '*' || config.nargs === '?' || config.const !== undefined || (positional && !required);
        var valuesCount = typeof config.nargs === 'number' ? config.nargs : undefined;
        return { positional: positional, multiple: multiple, required: required, allowEmpty: allowEmpty, valuesCount: valuesCount };
    };
    return ArgsParser;
}());
exports.ArgsParser = ArgsParser;
//# sourceMappingURL=classes.js.map